/*
 *  GNOM - packet radio-terminal for Linux
 *  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 

/****************************************************************************/
/**  Kommunikation zwischen GNOM und den dazugehrigen Kanaltreibern       **/
/****************************************************************************/

#include "global.h"
#include "comm.h"

#include "keyboard.h"
#include "toolbox.h"
#include "speaker.h"
#include "comp.h"
#include "display.h"
#include "utils.h"
#include "convert.h"
#include "sndqueue.h"
#include "savebuf.h"

#include <string.h>


/* System-Meldungen */
static char conn_str[] = "CONNECTED to";


/************************************************************************************************/

static void look_for_huffman(char *str, int *len)
{
  char huffstr[(MAXFRAMELEN*3)+1];
  char output[(MAXFRAMELEN*3)+1];
  
  memmove(huffstr+1,str,*len);
  huffstr[0] = (char)*len-1;
  if (!decstathuf(huffstr,output))
  {
  
syslog(LOG_DEBUG,"output[0]: %i, str[0]: %i output: %s",output[0],str[0],output);

    if (output[0] == str[0])
    {
      *len = (int)((uchar)output[0]+1);
      memcpy(str,output+1,*len);
    }
  }
}



/*
 * Trgt die Informationen eines Treibers in die Treiberliste ein. Die
 * Port-Liste bleibt noch leer.
 *
 * Gibt einen Pointer auf die Hardware-Struktur zurck.
 */
s_hardware *comm_add_driver(s_driver_infos_allg infos_allg, int fd)
{
  s_hardware *tmp;

  if (hardware == NULL)
  {
    /* Erster Eintrag */
    hardware = (s_hardware *)malloc(sizeof(s_hardware));
    tmp = hardware;
  }
  else {
    tmp = hardware;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_hardware *)malloc(sizeof(s_hardware));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  strcpy(tmp->name,infos_allg.name);
  strcpy(tmp->version,infos_allg.version);
  strcpy(tmp->autor,infos_allg.autor);
  tmp->ports = NULL;	/* Noch keine Ports definiert */
  tmp->fd = fd;
  
  if (fd+1 > maxfd) maxfd = fd+1;
  return(tmp);
}



/*
 * Trgt diesen Port in die bergebene Hardware-Struktur ein.
 */
void comm_add_driverport(s_hardware *hard, s_driver_infos_port portinfo)
{
  s_port_info *tmp;
  
  if (hard->ports == NULL)
  {
    /* Erster Eintrag */
    hard->ports = (s_port_info *)malloc(sizeof(s_port_info));
    tmp = hard->ports;
  }
  else {
    tmp = hard->ports;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_port_info *)malloc(sizeof(s_port_info));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  strcpy(tmp->port,portinfo.port);
  strcpy(tmp->desc,portinfo.desc);
  tmp->baud = portinfo.baud;
}



/*
 * Diese Funktion initialisiert die Treiber. Sie klappert alle sockets ab, die in
 * gnom.conf unter driver_sockets angegeben wurden, und holt sich da Konfigurations-
 * Infos 'raus.
 */
void comm_init_drivers()
{
  char *str,tmp[500];
  int i,count,fd,found;
  char anzports;
  s_driver_infos_port infos_port;
  s_driver_infos_allg infos_allg;
  int servlen;
  struct sockaddr_un serv_addr;
  fd_set fd_mask;
  s_hardware *hwptr;
  s_port_info *ports;
  uchar cmd;
  
  
  str = (char *)strdup(config->driver_sockets);
  hardware = NULL;
  
  /* Die einzelnen Sockets sind mit Komma getrennt */
  while (str[0] != '\0')
  {
    if ((i = POS(',',str)) != -1)
    {
      COPY(tmp,str,0,i);
      COPY(str,str,i+1,strlen(str)-i-1);
    }
    else
    {
      strcpy(tmp,str);
      str[0] = '\0';
    }
    
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sun_family = AF_UNIX;
    strcpy(serv_addr.sun_path,tmp);
    servlen = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
    /* open socket */
    if ((fd = socket(AF_UNIX,SOCK_STREAM,0)) < 0)
    {
      printf("ERROR! Cannot open socket (%s).\n",tmp);
      exit(1);
      return;
    }
    
    /* connect other program */
    if (connect(fd, (struct sockaddr *) &serv_addr, servlen) < 0)
    {
      close(fd);
      printf("ERROR! Cannot connect (%s).\n",tmp);
      exit(1);
    }
    
    /* Ersten Befehl abschicken */
    cmd = 1;
    write(fd,&cmd,1);
    
    /* Auf Antwort warten */
    FD_ZERO(&fd_mask);
    FD_SET(fd,&fd_mask);
    select(fd+1,&fd_mask,NULL,NULL,NULL);
    
    count = read(fd,&infos_allg,sizeof(s_driver_infos_allg));
    if (count != sizeof(s_driver_infos_allg))
    {
      close(fd);
      printf("ERROR! Cannot get port-infos (%s).\n",tmp);
      exit(1);
    }
    
    if (infos_allg.comm_version != COMM_VERSION)
       printf("  cannot use driver %s - wrong comm-protocol-version\n",infos_allg.name);
    else
    {
      /* Diesen Treiber in die Treiber-Liste eintragen. */
      hwptr = (s_hardware *) comm_add_driver(infos_allg,fd);
      printf("  driver found: %s V%s (%s)",infos_allg.name,infos_allg.version,infos_allg.autor);
    
      /* Auf Antwort warten */
      FD_ZERO(&fd_mask);
      FD_SET(fd,&fd_mask);
      select(fd+1,&fd_mask,NULL,NULL,NULL);
      count = read(fd,&anzports,1);
      if (count != 1)
      {
        close(fd);
        printf("ERROR! Cannot get port-infos (%s).\n",tmp);
        exit(1);
      }
      printf(" %i ports available:\n",anzports);

      i = 0;
      while (i < anzports)
      {
        FD_ZERO(&fd_mask);
        FD_SET(fd,&fd_mask);
        select(fd+1,&fd_mask,NULL,NULL,NULL);
      
        count = read(fd,&infos_port,sizeof(s_driver_infos_port));
        if (count != sizeof(s_driver_infos_port))
        {
          close(fd);
          printf("ERROR! Cannot get port-infos (%s).\n",tmp);
          exit(1);
        }
      
        /* Diesen Port in die Port-Liste eintragen */
        comm_add_driverport(hwptr,infos_port);
        printf("    port %s, %i baud (%s)\n",infos_port.port,infos_port.baud,infos_port.desc);
      
        i++;
      }
    }
  }
  
  
  /* Kontrolle, ob der in gnom.conf definierte Standard-Port auch vorhanden ist */
  hwptr = hardware;
  found = FALSE;
  while ((hwptr != NULL) && !found)
  {
    ports = hwptr->ports;
    while ((ports != NULL) && !found)
    {
      if (!strcmp(ports->port,config->stdport)) found = TRUE;
      ports = ports->next;
    }
    hwptr = hwptr->next;
  }
  
  if (!found)
  {
    printf("ERROR! Standardport (%s) is not defined in any driver\n",config->stdport);
    exit(1);
  }
  
  free(str);
}



/* ein Monitor-Frame kommt */
void comm_incoming_monitordata(s_hardware *hard)
{
  s_monitor moni;
  int count,len;
  char outtext[500],tmp[2000];
  
  
  if ((count = (read(hard->fd,&moni,sizeof(s_monitor)))) != sizeof(s_monitor)) return;
  
  /****************** Header des Frames **************/
  
  /* Portname */
  sprintf(outtext,"Port %s: ",moni.port);
  display_addline(outtext,strlen(outtext),MONICHAN,ATTR_MON_PORTNAME,FALSE);
  
  /* Protokoll */
  if ((moni.flags & MONI_AX_EAX) == 0)
     strcpy(outtext,"AX25: ");
  else
     strcpy(outtext,"EAX25: ");
  display_addline(outtext,strlen(outtext),MONICHAN,ATTR_MON_PROTOCOL,FALSE);
  
  if ((moni.flags & MONI_BAD_HEADER) == MONI_BAD_HEADER)
  {
    strcpy(outtext,"BAD HEADER");
    display_addline(outtext,strlen(outtext),MONICHAN,ATTR_MON_ADDR,FALSE);
  }
  else
  {
  
    /* Rufzeichen + Digis */
    display_addline(moni.calltocall,strlen(moni.calltocall),MONICHAN,ATTR_MON_ADDR,FALSE);
    if (moni.digis[0] != '\0')
    {
      strcpy(outtext," v ");
      display_addline(outtext,strlen(outtext),MONICHAN,ATTR_MON_AXHDR,FALSE);
      display_addline(moni.digis,strlen(moni.digis),MONICHAN,ATTR_MON_ADDR,FALSE);
    }
  
    switch (moni.frametype)
    {
      case MONI_SABM:
           strcpy(outtext," <SABM");
           break;
      case MONI_UA:
           strcpy(outtext," <UA");
           break;
      case MONI_I:
           strcpy(outtext," <I");
           break;
      case MONI_RR:
           strcpy(outtext," <RR");
           break;
      case MONI_RNR:
           strcpy(outtext," <RNR");
           break;
      case MONI_U:
           strcpy(outtext," <U");
           break;
      case MONI_DISC:
           strcpy(outtext," <DISC");
           break;
      case MONI_DM:
           strcpy(outtext," <DM");
           break;
      case MONI_FRMR:
           strcpy(outtext," <FRMR");
           break;
      case MONI_UI:
           strcpy(outtext," <UI");
           break;
      case MONI_SABME:
           strcpy(outtext," <SABME");
           break;
      case MONI_REJ:
           strcpy(outtext," <REJ");
           break;
      default: /* ERROR! */
               strcpy(outtext," <error");
               break;
    }
  
    switch (moni.cmd_resp)
    {
      case 1: /* COMMAND */
              strcat(outtext," C");
              if ((moni.flags & MONI_POLLFINAL) == MONI_POLLFINAL) strcat(outtext," P");
              break;
      case 2: /* RESPONSE */
              strcat(outtext," R");
              if ((moni.flags & MONI_POLLFINAL) == MONI_POLLFINAL) strcat(outtext," F");
              break;
      default: break;
    }
  
    if (moni.frametype == MONI_I)
    {
      sprintf(tmp," R%i S%i",moni.rxcount,moni.txcount);
      strcat(outtext,tmp);
    }
    if ((moni.frametype == MONI_RR) || (moni.frametype == MONI_REJ))
    {
      sprintf(tmp," R%i",moni.rxcount);
      strcat(outtext,tmp);
    }
  
    strcat(outtext,">");
    
    if (moni.frametype == MONI_FRMR)
    {
      strcat(outtext," ");
      strcat(outtext,moni.frmr_info);
    }
    
    /* PID */
    if ((moni.frametype == MONI_I) || (moni.frametype == MONI_UI))
    {
      strcat(outtext," pid=");
      switch (moni.pid)
      {
        case MONI_PID_SEGMENT:
             strcpy(tmp,"segment");
             break;
        case MONI_PID_ARP:
             strcpy(tmp,"ARP");
             break;
        case MONI_PID_NETROM:
             strcpy(tmp,"NETROM");
             break;
        case MONI_PID_IP:
             strcpy(tmp,"IP");
             break;
        case MONI_PID_X25:
             strcpy(tmp,"X.25");
             break;
        case MONI_PID_TEXNET:
             strcpy(tmp,"TEXNET");
             break;
        case MONI_PID_FLEXNET:
             strcpy(tmp,"FLEXNET");
             break;
        case MONI_PID_NO_L3:
             strcpy(tmp,"Text");
             break;
        default: sprintf(tmp,"%i",moni.pid);
      }
      strcat(outtext,tmp);
    }
  
    if ((moni.flags & MONI_DAMA) == MONI_DAMA)
       strcat(outtext," [DAMA]\xD");
    else
       strcat(outtext,"\xD");
  
    display_addline(outtext,strlen(outtext),MONICHAN,ATTR_MON_AXHDR,FALSE);
  }
  
  while (moni.anz_datalines > 0)
  {
    /* Lnge holen */
    if ((count = (read(hard->fd,&len,sizeof(len)))) != sizeof(len))
       display_addline("internal error!\xD",16,MONICHAN,ATTR_MON_DATA,FALSE);
    else
       if ((count = (read(hard->fd,&tmp,len))) != len)
          display_addline("internal error!\xD",16,MONICHAN,ATTR_MON_DATA,FALSE);
       else
       {
         if (tmp[len-1] != '\xD') tmp[len++] = '\xD';
         tmp[len] = '\0';
         /* String so umwandeln, da alle Bytes angezeigt werden knnen */
         checkshowchars(tmp,len);
         display_addline(tmp,len,MONICHAN,ATTR_MON_DATA,FALSE);
       }
    moni.anz_datalines--;
  }
  
  display_addline("\xD",1,MONICHAN,ATTR_MON_AXHDR,TRUE);
}



/* Von Auen kommt ein connect an */
void comm_incoming_connect(s_hardware *hard)
{
  int count,chan;
  s_got_connect coninfo;
  char tmp[500],out[500],tmp2[10],tmp3[10];


  if ((count = read(hard->fd,&coninfo,sizeof(s_got_connect))) != sizeof(s_got_connect)) return;
  
  if ((chan = chan_free(coninfo.mycall)) == -1) return;
  
  strcpy(chaninfo[chan]->call,coninfo.call);
  strcpy(chaninfo[chan]->mycall,coninfo.mycall);
  strcpy(chaninfo[chan]->port,coninfo.port);
  chaninfo[chan]->digis = (char *)strdup(coninfo.path);
  chaninfo[chan]->qsonr = coninfo.qsonr;
  chaninfo[chan]->driverfd = hard->fd;
  chaninfo[chan]->flags &= ~CH_CAN_SEND;	/* Im Augenblick darf nix gesendet werden */
  chaninfo[chan]->pwinfo = NULL;
  
  
  if (AktChan == chan) display_newscreen();
  chaninfo[chan]->status = 4;
  chaninfo[chan]->logintime = time(NULL);
  
  get_userdata_chan(chan);
  
  /* Connect-Info ausgeben */
  if (chan != AktChan) {
    chaninfo[chan]->flags |= CH_NEWDATA;
    update_lower_statusline(TRUE);
  }
  
  strcpy(chaninfo[chan]->firstcall,chaninfo[chan]->call);
  sprintf(tmp,"*** (%s) %s %s%s\r",
              chaninfo[chan]->port,conn_str,chaninfo[chan]->firstcall,chaninfo[chan]->digis);
  display_addline(tmp,strlen(tmp),chan,ATTR_RX_STATUS,TRUE);
  
  /* Nur den connect-Text schicken, wenns noch nicht passiert ist */
  if (!coninfo.old)
  {
    /* Connect-Text abschicken */
    
    /* Existiert im userdb-Verzeichnis ein File namens <call>.ctext? Wenn ja wird dieses als
     * ctext ausgesendet.
     */
    COPY(tmp2,chaninfo[chan]->firstcall,0,3);
    Klein(tmp2);
    strcpy(tmp3,chaninfo[chan]->firstcall);
    Klein(tmp3);
    sprintf(tmp,"%s/userdb/%s.../%s.ctext",config->maindir,tmp2,tmp3);
    if (!file_exist(tmp)) sprintf(tmp,"%s/ctext.gnom",config->maindir);
    send_macrofile(tmp,chan);
  }
  
  if (awaytext != NULL)
  {
    macro_thisline(tmp,awaytext,chan);
    sprintf(out,"## %s ##\r\r",tmp);
    send_Textline(chan,out);
  }
  
  /* Wenn fr dieses MYCALL kein anderer nicht connecteter Port mehr vorhanden ist,
   * dem Treiber bescheid sagen, da auf dieses Call nicht mehr gelauscht werden soll.
   */
  check_for_mycall_connect(chaninfo[chan]->mycall);
  
  todo_on_connect(chan);
  update_upper_statusline(TRUE);
  
  /* Erster Connect - Bimmeln */
  if (config->tones->con != NULL)
     if (config->tones->bell) speaker_playstring(config->tones->con);
}



int comm_which_channel(int driverfd, int qsonr)
{
  int i=1;
  
  while (i <= chancount)
  {
    if (chaninfo[i]->driverfd == driverfd)
       if (chaninfo[i]->qsonr == qsonr)
          return(i);
       i++;
  }
  
  return(-1);
}


/*
 * Von auen will jemand disconnecten
 */
void comm_incoming_disconect(s_hardware *hard)
{
  int count;
  int qsonr;
  int chan;
  
  if ((count = read(hard->fd,&qsonr,sizeof(qsonr))) != sizeof(qsonr)) return;
  
  if ((chan = comm_which_channel(hard->fd,qsonr)) == -1) return;
  
  disc_channel(chan);
}



/*
 * Hier kommt die Antwort auf einen Connect-Befehl
 */
void comm_incomming_connectanswer(s_hardware *hard)
{
  char code;
  int chan;
  int qsonr;
  
  read(hard->fd,&code,1);
  read(hard->fd,&chan,sizeof(chan));
  read(hard->fd,&qsonr,sizeof(qsonr));
  
  switch ((uchar)code)
  {
    case 1: ErrorText("INVALID PARAMETER",TRUE);
            disc_channel(chan);
            break;
    case 2: ErrorText("cannot open socket",TRUE);
            disc_channel(chan);
            break;
    case 3: ErrorText("cannot bind socket",TRUE);
            disc_channel(chan);
            break;
    case 4: ErrorText("cannot connect twice",TRUE);
            disc_channel(chan);
            break;
    case 5: ErrorText("unknown port",TRUE);
            disc_channel(chan);
            break;
    case 0xFF: /* Ok, der Connect wird aufgebaut */
               chaninfo[chan]->qsonr = qsonr;
               break;
  }
}



/*
 * Ankommende Daten
 */
void comm_incoming_data(s_hardware *hard)
{
  int len;
  int qsonr,chan;
  char *str,*showstr;
  
  /* Zu welchem qso gehren die daten? */
  read(hard->fd,&qsonr,sizeof(qsonr));
  chan = comm_which_channel(hard->fd,qsonr);
  
  /* Jetzt die Lnge holen */
  read(hard->fd,&len,sizeof(len));
  
  str = (char *)malloc((len*3)+1);
  read(hard->fd,str,len);
  str[len] = '\0';
  
  /* Ist dieses Frame Huffman-Gepackt? */
  if ((chaninfo[chan]->flags & CH_COMP) == CH_COMP)
      look_for_huffman(str,&len);
  
  /* Umlautwandlung */
  if (((chaninfo[chan]->flags & CH_AUTOBINRX) == 0) && ((chaninfo[chan]->flags & CH_7PLUS) == 0) &&
      ((chaninfo[chan]->flags & CH_AUTOBINTX) == 0))
           conv_it_rx(str,len,chaninfo[chan]->convert_rx);
  
  if (((chaninfo[chan]->flags & CH_AUTOBINRX) == 0) && ((chaninfo[chan]->flags & CH_AUTOBINTX) == 0) &&
      ((chaninfo[chan]->flags & CH_7PLUS) == 0))
  {

    showstr = (char *)malloc(len);
    memcpy(showstr,str,len);
    checkshowchars(showstr,len);
    display_addline(showstr,len,chan,ATTR_RX_EMPF,TRUE);
    free(showstr);
    
    if (config->tones->klack) speaker_klack();
    /* Wenn dieser Kanal nicht als der Aktuelle ausgewhlt wurde, und was 'reinkommt,
     * wird der entsprechende Kanal als "neue Daten" markiert.
     * (An dieser Stelle, damit das display_newscreen vom display_addline mit
     * verwendet werden kann.)
     */
    if (AktChan != chan)
    {
      chaninfo[chan]->flags |= CH_NEWDATA;
      update_lower_statusline(TRUE);
    }
  }

  /* Speichert das Frame unter chaninfo[i]->rxframes */
  data_save_frame(str,len,chan);

  /* Anzahl empfangener Bytes auf diesem Kanal hochzhlen */
  chaninfo[chan]->rx_bytes += len;

  /* Die einzelnen Frames an die Subroutinen verteilen */
  data_split_frames(chan);

  free(str);
}




/*
 * Irgendwas in den /proc/net/ax25 hat sich gendert. Rausfinden was und
 * entsprechend reagieren
 */
void comm_incoming_status(s_hardware *hard)
{
  char tmp[2];
  char str[100];
  int chan,qsonr;
  
  read(hard->fd,&tmp,2);
  read(hard->fd,&qsonr,sizeof(qsonr));
  chan = comm_which_channel(hard->fd,qsonr);

  switch (tmp[0])
  {
    case POLL_STATUS: chaninfo[chan]->oldstatus = chaninfo[chan]->status;
                      /* Haben wir connected? */
                      if ((chaninfo[chan]->oldstatus == 1) && ((int)tmp[1] == 4))
                      {
                        if (chan != AktChan)
                        {
                          chaninfo[chan]->flags |= CH_NEWDATA;
                          update_lower_statusline(TRUE);
                        }
                        strcpy(chaninfo[chan]->firstcall,chaninfo[chan]->call);
                        
                        chaninfo[chan]->pwinfo = NULL;
                        todo_on_connect(chan);
                        
                        sprintf(str,"*** (%s) %s %s%s\r",
                                     chaninfo[chan]->port,conn_str,chaninfo[chan]->firstcall,chaninfo[chan]->digis);
                        display_addline(str,strlen(str),chan,ATTR_RX_STATUS,TRUE);
                        
                        /* Erster Connect - Bimmeln */
                        if (config->tones->con != NULL)
                           if (config->tones->bell) speaker_playstring(config->tones->con);
                      }
                      
                      /* Verbindungsabbau */
                      if ((chaninfo[chan]->oldstatus > 1) && ((int)tmp[1] == 0))
                         disc_channel(chan);
                      chaninfo[chan]->status = tmp[1];
                      update_upper_statusline(TRUE);
                      break;
    case POLL_TRIES: chaninfo[chan]->tries = (int)tmp[1];
                     update_upper_statusline(TRUE);
                     break;
    case POLL_SQUEUE: chaninfo[chan]->squeue = (int)tmp[1];
                      update_upper_statusline(TRUE);
                      break;
    case POLL_UNACK: chaninfo[chan]->unack = (int)tmp[1];
                     update_upper_statusline(TRUE);
                     break;
  }
}



/*
 * Alle Kanle connecten, die diesen Treiber benutzt haben
 */
void comm_disconnect_this_driver(s_hardware *hard)
{
  int i=1;
  
  while (i <= chancount)
  {
    if (chaninfo[i]->driverfd == hard->fd)
    {
      chaninfo[i]->qsonr = -1;
      disc_channel(i);
    }
    i++;
  }
}



/*
 * Der Treiber hat das empfangene Frame verstanden und hat noch Platz
 * fr ein weiteres
 */
void comm_incoming_send_next_frame(s_hardware *hard)
{
  int count,chan,qsonr;
  
  if ((count = read(hard->fd,&qsonr,sizeof(qsonr))) != sizeof(qsonr)) return;
  
  if ((chan = comm_which_channel(hard->fd,qsonr)) == -1) return;
  
  chaninfo[chan]->flags |= CH_CAN_SEND;
  send_a_frame(chan);
}



/*
 * Der Treiber kann keine weiteren Daten mehr verarbeiten. Also schicken
 * wir ihm keine mehr. (Flag auf 0 setzen)
 */
void comm_incoming_no_more_frames(s_hardware *hard)
{
  int count,chan,qsonr;
  
  if ((count = read(hard->fd,&qsonr,sizeof(qsonr))) != sizeof(qsonr)) return;
  
  if ((chan = comm_which_channel(hard->fd,qsonr)) == -1) return;
  
  chaninfo[chan]->flags &= ~CH_CAN_SEND;
}



/*
 * Dieser Treiber will was von uns
 */
void comm_driver_command(s_hardware *hard)
{
  int count;
  char cmd;
  
  if ((count = read(hard->fd,&cmd,1)) == 0)
  {
    /* Der Treiber wurde beendet */
    comm_disconnect_this_driver(hard);
    delete_this_driver_from_list(hard);
    return;
  }
  
  /* Daten vom Treiber. */
  switch (cmd)
  {
    case 2:  /* Antwort auf einen Connect-Befehl */
             comm_incomming_connectanswer(hard);
             break;
    case 3:  /* Ein Connect von auen */
             comm_incoming_connect(hard);
             break;
    case 4:  /* Status-nderungen */
       	     comm_incoming_status(hard);
             break;
    case 5:  /* Disconnect */
             comm_incoming_disconect(hard);
             break;
    case 6:  /* Daten empfangen */
             comm_incoming_data(hard);
             break;
    case 7:  /* Alle noch gespeicherten QSO-Daten wurden vom Treiber abgeschickt. */
             restore_channelstatus();
             update_upper_statusline(TRUE);
             update_lower_statusline(TRUE);
             break;
    case 40: /* Das nchste Frame darf gesendet werden */
             comm_incoming_send_next_frame(hard);
             break;
    case 41: /* Jetzt darf nix mehr gesendet werden */
             comm_incoming_no_more_frames(hard);
             break;
    case 50: comm_incoming_monitordata(hard);
             break;
  }
}



void comm_close_connection(int chan)
{
  char cmd=5;
  
  write(chaninfo[chan]->driverfd,&cmd,1);
  write(chaninfo[chan]->driverfd,&chaninfo[chan]->qsonr,sizeof(chaninfo[chan]->qsonr));
}



void comm_make_connect(int chan, char *call, char *path, char *mycall, char *port, int driverfd)
{
  char cmd=2;
  s_make_connect coninf;
  
  strcpy(coninf.port,port);
  strcpy(coninf.mycall,mycall);
  strcpy(coninf.call,call);
  strcpy(coninf.digis,path);
  coninf.chan = chan;
  
  write(chaninfo[chan]->driverfd,&cmd,1);
  write(driverfd,&coninf,sizeof(s_make_connect));
}



void comm_send_frame(int chan, char *str, int len)
{
  char cmd=6;

  /* Anzahl tx_bytes auf diesem Kanal hochzhlen */
  chaninfo[chan]->tx_bytes += len;
  
  write(chaninfo[chan]->driverfd,&cmd,1);
  write(chaninfo[chan]->driverfd,&chaninfo[chan]->qsonr,sizeof(chaninfo[chan]->qsonr));
  write(chaninfo[chan]->driverfd,&len,sizeof(len));
  write(chaninfo[chan]->driverfd,str,len);
}



/*
 * Auf dieses Rufzeichen auf diesem Port lauschen
 */
void comm_listen_to(char *call, s_port_info *port, int fd)
{
  s_listen_to listen;
  char cmd=7;
  strcpy(listen.call,call);
  strcpy(listen.port,port->port);
  
  write(fd,&cmd,1);
  write(fd,&listen,sizeof(s_listen_to));
}



/*
 * Auf dieses Rufzeichen auf diesem Port nicht mehr lauschen
 */
void comm_dont_listen_to(char *call, s_port_info *port, int fd)
{
  s_listen_to listen;
  char cmd=8;
  
  strcpy(listen.call,call);
  strcpy(listen.port,port->port);
  
  write(fd,&cmd,1);
  write(fd,&listen,sizeof(s_listen_to));
}


void comm_monitor_activate()
{
  int cmd=50;
  s_hardware *hard;

  hard = hardware;
  while (hard != NULL)
  {
    write(hard->fd,&cmd,1);
    hard = hard->next;
  }
}


void comm_monitor_deactivate()
{
  int cmd=51;
  s_hardware *hard;

  hard = hardware;
  while (hard != NULL)
  {
    write(hard->fd,&cmd,1);
    hard = hard->next;
  }
}



