/*
 *  GNOM - packet radio-terminal for Linux
 *  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/****************************************************************************/
/**  Die Termial-Befehle                                                   **/
/****************************************************************************/


#include "global.h"
#include "commands.h"

#include "comm.h"
#include "keyboard.h"
#include "display.h"
#include "utils.h"
#include "toolbox.h"
#include "password.h"
#include "sndqueue.h"
#include "dostime.h"
#include "remote.h"

#include <string.h>

static char invalidcall_txt[] = "INVALID CALLSIGN";
static char alreadyconn_str[] = "CHANNEL ALREADY CONNECTED";


void cmd_Quit()
{
  char quittext[] = "Quit? (Y/N)";

  ErrorText(quittext,FALSE);
  wmove(tx_win,y_akt,strlen(quittext));
  wrefresh(tx_win);
  ConsoleStatus = 1;		/* Auf Antwort beim QUIT-Befehl warten */
}



/* Sendet ein ASCII-File aus */
void cmd_Send(char *params)
{
  char tmp[500];

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (!file_exist(params))
  {
    /* File existiert nicht - Fehlermeldung */
    sprintf(tmp,"Error, file '%s' not found.",params);
    ErrorText(tmp,TRUE);
    wrefresh(tx_win);
  }
  else
  {
    ErrorText("Ok, sent.",TRUE);
    send_Plainfile(AktChan,params);
  }

  wrefresh(tx_win);
}



void cmd_disc()
{

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("CHANNEL NOT CONNECTED",TRUE);
    return;
  }

  disc_channel(AktChan);

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  ErrorText("Ok",TRUE);

}



void cmd_name(char *name)
{
  char tmp[100];

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected!",TRUE);
    return;
  }

  setname(chaninfo[AktChan]->call,name);

  if (chaninfo[AktChan]->name != NULL) free(chaninfo[AktChan]->name);
  chaninfo[AktChan]->name = (char *)strdup(name);

  sprintf(tmp,"OK, name set to '%s'.",name);
  ErrorText(tmp,TRUE);
  update_upper_statusline(TRUE);
}



void cmd_connect(char *call)
{
  int i,ok,callcount;
  char port[30],tmp[100],tmp2[100],usethis[100],onlyfirst[10],digis[100];
  
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }
  
  if (chaninfo[AktChan]->driverfd != -1)
  {
    ErrorText(alreadyconn_str,TRUE);
    return;
  }
  
  
  if (hardware == NULL)
  {
    ErrorText("Sri, no hardware is attached. Please restart the Program.",TRUE);
    return;
  }
  
  if (call[0] == '\0')
  {
    ErrorText("INVALID CALLSIGN",TRUE);
    return;
  }
  
  strcpy(port,config->stdport);
  
  /* Gucken, ob die Rufzeichen auch korrekt sind (nicht zu lang, SSID nicht ber 15 etc) */
  strcpy(tmp,call);
  KillSpacesRight(tmp);
  KillSpacesLeft(tmp);
  OnlyOneSpace(tmp2,tmp);
  strcpy(tmp,tmp2);
  ok = FALSE;
  callcount = 0;
  digis[0] = '\0';
  while (!ok)
  {
    i = POS(' ',tmp);
    if (i != -1)
    {
      COPY(tmp2,tmp,0,i);
      COPY(tmp,tmp,i+1,strlen(tmp)-i-1);
    }
    else {
      ok = TRUE;
      strcpy(tmp2,tmp);
    }
    callcount++;
    
    if (callcount == 1)
    {
      if (valid_portname(tmp2))
      {
        strcpy(port,tmp2);
        callcount--;
      }
      else
      {
        Gross(tmp2);
        Gross(tmp);
        Gross(call);
        strcpy(usethis,tmp2);
        strcpy(onlyfirst,tmp2);
      }
    }
    else if (callcount == 2) {
           sprintf(usethis,"%s via %s",usethis,tmp2);
           sprintf(digis," via %s",tmp2);
         }
         else
         {
           sprintf(usethis,"%s %s",usethis,tmp2);
           sprintf(digis,"%s %s",digis,tmp2);
         }

    if (!valid_callsign(tmp2))
    {
      ErrorText(invalidcall_txt,TRUE);
      return;
    }
  }
  
  
  chaninfo[AktChan]->driverfd = ax25_get_driverfd(port);
  
  comm_make_connect(AktChan,onlyfirst,digis,chaninfo[AktChan]->mycall,port,chaninfo[AktChan]->driverfd);
  
  chaninfo[AktChan]->status = 1;
  strcpy(chaninfo[AktChan]->port,port);
  strcpy(chaninfo[AktChan]->call,onlyfirst);
  strcpy(chaninfo[AktChan]->firstcall,onlyfirst);
  chaninfo[AktChan]->flags &= ~CH_CAN_SEND;		/* Im Augenblick darf nix gesendet werden */
  chaninfo[AktChan]->digis = (char *)strdup(digis);
  chaninfo[AktChan]->pwinfo = NULL;
  get_userdata_chan(AktChan);

  check_for_mycall_connect(chaninfo[AktChan]->mycall);
  
  display_newscreen();
}



void cmd_remote(char *param)
{
  int rebuild = FALSE;
  
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (param[0] == '\0')
  {
    /* Keinen Parameter angegeben - aktuellen Status ausgeben */
    if ((chaninfo[AktChan]->flags & CH_REMOTE) == CH_REMOTE)
       ErrorText("ON",TRUE);
    else
       ErrorText("OFF",TRUE);
    return;
  }
  
  Gross(param);
  
  if (!strcmp(param,"ON") || (param[0] == 'Y') || (param[0] == '1'))
  {
    ErrorText("Ok",TRUE);
    if ((chaninfo[AktChan]->flags & CH_REMOTE) == 0) rebuild = TRUE;
    chaninfo[AktChan]->flags |= CH_REMOTE;
    if (rebuild) update_upper_statusline(TRUE);
    return;
  }
  
  if (!strcmp(param,"OFF") || (param[0] == 'N') || (param[0] == '0'))
  {
    ErrorText("Ok",TRUE);
    if ((chaninfo[AktChan]->flags & CH_REMOTE) == CH_REMOTE) rebuild = TRUE;
    chaninfo[AktChan]->flags &= ~CH_REMOTE;
    if (rebuild) update_upper_statusline(TRUE);
    return;
  }
  
  ErrorText("Wrong value.",TRUE);
}



void cmd_mycall(char *param)
{
  char tmp[100];
  
  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }
  
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  if (chaninfo[AktChan]->driverfd != -1)
  {
    ErrorText("Not while connected!",TRUE);
    return;
  }
  
  if (param[0] == '\0')
  {
    /* Aktuelles MYCALL ausgeben */
    ErrorText(chaninfo[AktChan]->mycall,TRUE);
    return;
  }
  
  strcpy(tmp,param);
  Gross(tmp);
  if (valid_callsign(tmp))
  {
    strcpy(chaninfo[AktChan]->mycall,tmp);
    mycall_addcall(tmp);
    ErrorText("Ok",TRUE);
  }
  else ErrorText(invalidcall_txt,TRUE);
}



/* Schreibt ein File auf die eigene Platte */
void cmd_save(char *name)
{
  int fd,i;
  char tmp[500];
  char filename[500];
  
  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (name[0] == '\0')
  {
    ErrorText("Syntax: SAVE <filename>",TRUE);
    return;
  }

  if ((chaninfo[AktChan]->flags & CH_SAVE) == CH_SAVE)
  {
    ErrorText("A file is open. Please close it and try again.",TRUE);
    return;
  }
  if ((chaninfo[AktChan]->flags & CH_7PLUS) == CH_7PLUS)
  {
    ErrorText("A 7plus-file is open. Please close it and try again.",TRUE);
    return;
  }
  if ((chaninfo[AktChan]->flags & CH_AUTOBINRX) == CH_AUTOBINRX)
  {
    ErrorText("A AutoBIN-file is open. Please close it and try again.",TRUE);
    return;
  }
  if ((chaninfo[AktChan]->flags & CH_AUTOBINTX) == CH_AUTOBINTX)
  {
    ErrorText("A AutoBIN-file is being transmitted. Please close it and try again.",TRUE);
    return;
  }
  
  /* Gucken, ob ein Verzeichnis mit angegeben wurde. Wenn nein wird das File
   * automatisch ins "down"-Verzeichnis gepackt.
   */
  if ((i = POS('/',name)) == -1)
     sprintf(filename,"%s/down/%s",config->maindir,name);
  else
     strcpy(filename,name);
  
  if ((fd = open(filename,O_RDWR)) != -1)
  {
    /* File existiert schon - anhngen */
    ErrorText("File already exists - appending...",TRUE);
    close(fd);
    fd = open(filename,O_WRONLY|O_APPEND);
  }
  else
  {
    /* File existiert noch nicht - probieren, ob wirs neu anlegen knnen */
    if ((fd = open(filename,O_WRONLY|O_CREAT)) == -1)
    {
      /* Nein, knnen wir nicht - Fehlermeldung */
      sprintf(tmp,"Error! Cannot open %s.",filename);
      ErrorText(tmp,TRUE);
      return;
    }
  }
  
  sprintf(tmp,"File '%s' is open now.",filename);
  ErrorText(tmp,TRUE);
  
  /* die chaninfo[]->saveplain - struktur allokieren etc */
  chaninfo[AktChan]->saveplain = (s_saveplain *)malloc(sizeof(s_saveplain));
  chaninfo[AktChan]->saveplain->filename = (char *)strdup(filename);
  chaninfo[AktChan]->saveplain->local = 1;
  chaninfo[AktChan]->saveplain->fd = fd;
  
  /* Zugriffsrechte einstellen */
  fchmod(fd,S_IRUSR|S_IWUSR);
  
  chaninfo[AktChan]->flags |= CH_SAVE;
  update_upper_statusline(TRUE);
}



/* Schliet eine Speicherung auf dem aktuellen Kanal. Entweder saveplaintext, 7plus, autobin oder
 * irgendwas anderes
 */
void cmd_close()
{
  char tmp[500];
  
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }
  
  /* Ist ein normaler Text-Save offen? */
  if ((chaninfo[AktChan]->flags & CH_SAVE) == CH_SAVE)
  {
    if (chaninfo[AktChan]->saveplain->local == 2)
    {
      sprintf(tmp,"<GNOM>: '%s' closed by SysOp.\r",chaninfo[AktChan]->saveplain->filename);
      send_Textline(AktChan,tmp);
    }
    sprintf(tmp,"OK, '%s' closed.",chaninfo[AktChan]->saveplain->filename);
    ErrorText(tmp,TRUE);
    close(chaninfo[AktChan]->saveplain->fd);
    free(chaninfo[AktChan]->saveplain->filename);
    free(chaninfo[AktChan]->saveplain);
    chaninfo[AktChan]->saveplain = NULL;
  
    chaninfo[AktChan]->flags &= ~CH_SAVE;
    update_upper_statusline(TRUE);
    return;
  }
  
  if ((chaninfo[AktChan]->flags & CH_7PLUS) == CH_7PLUS)
  {
    /* Das wars, das 7plus-File ist zu Ende */
    close(chaninfo[AktChan]->auto7plus->fd);
    strcpy(tmp,"<GNOM>: 7plus-Autosave aborted by SysOp.\r");
    
    if (chaninfo[AktChan]->type == TYPE_TERMINAL)
       send_Textline(AktChan,tmp);
    else
       display_addline(tmp,strlen(tmp),AktChan,ATTR_RX_SEND,TRUE);
    
    free(chaninfo[AktChan]->auto7plus->name);
    free(chaninfo[AktChan]->auto7plus);
    chaninfo[AktChan]->auto7plus = NULL;
    
    chaninfo[AktChan]->flags &= ~CH_7PLUS;
    update_lower_statusline(TRUE);
    update_upper_statusline(TRUE);
  }
  
}



/* Sendet ein AutoBIN-File aus */
void cmd_Send_AutoBIN(char *params)
{
  char tmp[500],tmp2[500];
  unsigned short crc;
  long size;
  int i;
  struct stat stat_data;
  unsigned short time,date;
  unsigned long fulldostime;


  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected.",TRUE);
    return;
  }


  if ((chaninfo[AktChan]->flags & CH_AUTOBINRX) == CH_AUTOBINRX)
  {
    ErrorText("Close the AutoBIN-RX first.",TRUE);
    return;
  }


  if (!file_exist(params))
  {
    /* File existiert nicht - Fehlermeldung */
    sprintf(tmp,"Error, file '%s' not found.",params);
    ErrorText(tmp,TRUE);
    wrefresh(tx_win);
    return;
  }

  crc = calc_crctph_wholefile(params);
  size = filesize(params);

  ErrorText("Ok, sending.",TRUE);
/*  send_Plainfile(AktChan,params);*/

  /* Dateiname ohne Pfad */
  strcpy(tmp2,params);
  while ((i = POS('/',tmp2)) != -1) COPY(tmp2,tmp2,i+1,strlen(tmp2)-i-1); 
  while ((i = POS('\\',tmp2)) != -1) COPY(tmp2,tmp2,i+1,strlen(tmp2)-i-1);

  /* Datenstruktur */
  chaninfo[AktChan]->autobin = (s_autobin *)malloc(sizeof(s_autobin));
  chaninfo[AktChan]->autobin->fd = -1;
  chaninfo[AktChan]->autobin->name = (char *)strdup(params);
  chaninfo[AktChan]->autobin->len = size;
  chaninfo[AktChan]->autobin->bytes_got = 0;
  chaninfo[AktChan]->autobin->calc_crc = 1;
  chaninfo[AktChan]->autobin->crc_rx = 0;
  chaninfo[AktChan]->autobin->crc_tx = crc;

  /* File-Datum */
  stat(params,&stat_data);
  date_unix2dos(stat_data.st_mtime,&time,&date);
  fulldostime = (date << 16) + time;

  /* #EBIN# - Zeile absenden */
/*  sprintf(tmp,"\xD#EBIN#0%.4lX#1%.2X%s#2%.4X#4%.8lX\xD",
             size,strlen(chaninfo[AktChan]->autobin->name),chaninfo[AktChan]->autobin->name,crc,stat_data.st_mtime);
  send_Textline(AktChan,tmp);*/

  /* Normaler #BIN# - Header */  
  sprintf(tmp,"\xD#BIN#%li#|%i#$%lX#%s\xD",size,crc,fulldostime,tmp2);
  send_Textline(AktChan,tmp);
  
  chaninfo[AktChan]->consstat = 1;
  
  wrefresh(tx_win);
}



/* Passwort */
void cmd_sys()
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected!",TRUE);
    return;
  }

  /* Keine Passwort-Beantwortung whrend AutoBIN */
  if (((chaninfo[AktChan]->flags & CH_AUTOBINRX) == CH_AUTOBINRX) ||
      ((chaninfo[AktChan]->flags & CH_AUTOBINTX) == CH_AUTOBINTX))
  {
    ErrorText("No passwort-computing while AutoBIN-transfers.",TRUE);
    return;
  }
  
  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected.",TRUE);
    return;
  }

  password(AktChan);

}



void cmd_settype(char *params)
{
  int type;


  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected.",TRUE);
    return;
  }

  if (params[0] == '\0')
  {
    switch (chaninfo[AktChan]->type)
    {
      case TYPE_TERMINAL: ErrorText("Type is 'terminal'.",TRUE); break;
      case TYPE_MAILBOX: ErrorText("Type is 'mailbox'.",TRUE); break;
      case TYPE_DIGI: ErrorText("Type is 'digipeater'.",TRUE); break;
      case TYPE_CONVERS: ErrorText("Type is 'convers'.",TRUE); break;
    }
    return;
  }


  type = atoi(params);
  
  if (type == 0)
  {
    Gross(params);
    if (!strcmp(params,"BOX"))
       type = TYPE_MAILBOX;
    else if (!strcmp(params,"DIGI"))
            type = TYPE_DIGI;
         else if (!strcmp(params,"USER"))
                 type = TYPE_TERMINAL;
              else if (!strcmp(params,"CONV"))
                      type = TYPE_CONVERS;
  }
  
  if ((type < 0) || (type > 4))
  {
    ErrorText("settype: type must be between 1 and 4.",TRUE);
    return;
  }

  settype(chaninfo[AktChan]->call,type);
  chaninfo[AktChan]->type = type;

  switch (type)
  {
    case TYPE_TERMINAL: ErrorText("Ok, type set to 'terminal'.",TRUE); break;
    case TYPE_MAILBOX: ErrorText("Ok, type set to 'mailbox'.",TRUE); break;
    case TYPE_DIGI: ErrorText("Ok, type set to 'digipeater'.",TRUE); break;
    case TYPE_CONVERS: ErrorText("Ok, type set to 'convers'.",TRUE); break;
  }
}



void cmd_comp(char *params)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected!",TRUE);
    return;
  }

  if (params[0] == '\0')
  {
    if ((chaninfo[AktChan]->flags & CH_COMP) == 0)
       ErrorText("deactivated",TRUE);
    else
       ErrorText("activated",TRUE);
    return;
  }

  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    chaninfo[AktChan]->flags |= CH_COMP;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }

  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    chaninfo[AktChan]->flags &= ~CH_COMP;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}



void cmd_autobin(char *params)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (params[0] == '\0')
  {
    if ((chaninfo[AktChan]->flags & CH_CHECKAUTOBIN) == 0)
       ErrorText("deactivated",TRUE);
    else
       ErrorText("activated",TRUE);
    return;
  }

  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    chaninfo[AktChan]->flags |= CH_CHECKAUTOBIN;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }

  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    chaninfo[AktChan]->flags &= ~CH_CHECKAUTOBIN;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}


void cmd_auto7plus(char *params)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (params[0] == '\0')
  {
    if ((chaninfo[AktChan]->flags & CH_CHECK7PLUS) == 0)
       ErrorText("deactivated",TRUE);
    else
       ErrorText("activated",TRUE);
    return;
  }

  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    chaninfo[AktChan]->flags |= CH_CHECK7PLUS;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }

  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    chaninfo[AktChan]->flags &= ~CH_CHECK7PLUS;
    ErrorText("Ok",TRUE);
    update_upper_statusline(TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}



void cmd_redraw()
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  /* Den kompletten Bildschirm lschen und wieder neu aufbauen */
  wclear(tx_win);
  wclear(rx_win);
  wclear(stoben_win);
  wclear(stunten_win);
  display_newscreen();

}



void cmd_cookie()
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  send_cookie(AktChan);
  
}



void cmd_framesize(char *param)
{
  char tmp[100];
  int i;

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);


  if (param[0] == '\0')
  {
    sprintf(tmp,"framesize: %i",framesize);
    ErrorText(tmp,TRUE);
    return;
  }

  i = atoi(param);
  
  if (i == 0)
  {
    sprintf(tmp,"wrong framesize: %s",param);
    ErrorText(tmp,TRUE);
    return;
  }
  
  framesize = i;
  
  sprintf(tmp,"Ok, framesize set to %i.",framesize);
  ErrorText(tmp,TRUE);
  
}



void cmd_test()
{
  int i,k=0;
  char str[300];

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  i = 33;
  while (i < 256)
  {
    str[k] = (char)i;
    i++;
    k++;
  }
  str[k] = '\0';

  display_addline(str,strlen(str),AktChan,ATTR_RX_SEND,TRUE);
  display_addline("\xD",1,AktChan,ATTR_RX_SEND,TRUE);
  
}



void cmd_ports()
{
  s_port_info *ports;
  char tmp[100];
  s_hardware *hard;
  
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  ErrorText("      driver         port     baud      desc  ",TRUE);
  ErrorText("-----------------------------------------------------",TRUE);
  
  /* Gibt Informationen ber alle vorhandenen Ports aus. */
  hard = hardware;
  while (hard != NULL)
  {
    ports = hard->ports;
    while (ports != NULL)
    {
       sprintf(tmp," %10s V%-4s  %-10s %5i %10s"
                  ,hard->name,hard->version
                  ,ports->port,ports->baud,ports->desc);
     
      ErrorText(tmp,TRUE);
      ports = ports->next;
    }
    hard = hard->next;
  }
}



/* Zeichensatz-Konvertierung */
void cmd_convert(char *param)
{
  int i;
  char tmp[100],tmp1[100],tmp2[100];
  int tx,rx;

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (param[0] == '\0')
  {
    ErrorText("Syntax: CONVERT <rx>/<tx>",TRUE);
    return;
  }
  
  KillAllSpaces(tmp,param);

  if ((i = POS('/',tmp)) == -1)
  {
    ErrorText("Syntax: CONVERT <rx>/<tx>",TRUE);
    return;
  }

  COPY(tmp1,tmp,0,i);
  COPY(tmp2,tmp,i+1,strlen(tmp)-i-1);
  
  rx = atoi(tmp1);
  tx = atoi(tmp2);

  if ((tx > 1) || (tx < 0) || (rx > 1) || (rx < 0))
  {
    ErrorText("Error! Wrong value.",TRUE);
    return;
  }
  
  chaninfo[AktChan]->convert_tx = tx;
  chaninfo[AktChan]->convert_rx = rx;
  
  update_upper_statusline(TRUE);
}



void cmd_cbell(char *param)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (param[0] == '\0')
  {
    if (config->tones->bell)
       ErrorText("activated",TRUE);
    else
       ErrorText("deactivated",TRUE);
    return;
  }
  
  Gross(param);
  if (!strcmp(param,"ON") || (param[0] == 'Y') || (param[0] == '1'))
  {
    config->tones->bell = TRUE;
    ErrorText("Ok",TRUE);
    return;
  }
  if (!strcmp(param,"OFF") || (param[0] == 'N') || (param[0] == '0'))
  {
    config->tones->bell = FALSE;
    ErrorText("Ok",TRUE);
    return;
  }
  
  ErrorText("Wrong parameter.",TRUE);
}



void cmd_ring(char *param)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (param[0] == '\0')
  {
    if (config->tones->ringok)
       ErrorText("activated",TRUE);
    else
       ErrorText("deactivated",TRUE);
    return;
  }
  
  Gross(param);
  if (!strcmp(param,"ON") || (param[0] == 'Y') || (param[0] == '1'))
  {
    config->tones->ringok = TRUE;
    ErrorText("Ok",TRUE);
    return;
  }
  if (!strcmp(param,"OFF") || (param[0] == 'N') || (param[0] == '0'))
  {
    config->tones->ringok = FALSE;
    ErrorText("Ok",TRUE);
    return;
  }
  
  ErrorText("Wrong parameter.",TRUE);
}



void cmd_rtt()
{
  char tmp[100];

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected!",TRUE);
    return;
  }
  
  sprintf(tmp,"//ECHO //RTT %lX\r",time(NULL));
  send_Textline(AktChan,tmp);
  
  ErrorText("Ok",TRUE);
}



void cmd_channel(char *param)
{
  int chan;


  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (param[0] == '\0')
  {
    ErrorText("Syntax: CHANNEL <chan-nr>",TRUE);
    return;
  }

  chan = atoi(param);
  
  if ((chan < 1) || (chan > chancount))
  {
    ErrorText("Invalid channel-number",TRUE);
    return;
  }
  
  /* Wenn wir eh schon auf diesem Kanal sind, nix machen */
  if (AktChan == chan)
  {
    ErrorText("Already on the choosen channel",TRUE);
    return;
  }
  
  /* Auf dem alten Kanal im Scrollbuffer ganz nach Unten gehen */
  while (chaninfo[AktChan]->scroll_screen != chaninfo[AktChan]->scroll_screen_upper)
  {
     chaninfo[AktChan]->scroll_screen = chaninfo[AktChan]->scroll_screen->next;
  }

  /* Position des Cursors im Vorschreib-Puffers speichern */
  chaninfo[AktChan]->x_akt = x_akt;
  chaninfo[AktChan]->y_akt = y_akt;

  /* Neuer Kanal */
  AktChan = chan;

  /* Dieser Kanal hat keine neuen Infos mehr, da wir ja jetzt gerade gucken */
  chaninfo[AktChan]->flags &= ~CH_NEWDATA;

  /* Position des Cursors neu setzen */
  x_akt = chaninfo[AktChan]->x_akt;
  y_akt = chaninfo[AktChan]->y_akt;

  display_newscreen();
}



/* Wechselt auf den ersten freien Kanal */
void cmd_freech()
{
  int i,found=FALSE;
  
  
  i = 1;
  while ((i <= chancount) && (!found))
  {
    if (chaninfo[i]->driverfd == -1)
       found = TRUE;
    else
       i++;
  }
  
  if (!found)
  {
    ErrorText("No free channel available",TRUE);
    return;
  }
  

  if (AktChan == i) return;		/* Wir sind auf dem ersten freien Kanal */
  
  /* Auf dem alten Kanal im Scrollbuffer ganz nach Unten gehen */
  while (chaninfo[AktChan]->scroll_screen != chaninfo[AktChan]->scroll_screen_upper)
  {
    chaninfo[AktChan]->scroll_screen = chaninfo[AktChan]->scroll_screen->next;
  }
  
  /* Position des Cursors im Vorschreib-Puffers speichern */
  chaninfo[AktChan]->x_akt = x_akt;
  chaninfo[AktChan]->y_akt = y_akt;
  
  /* Neuer Kanal */
  AktChan = i;

  /* Dieser Kanal hat keine neuen Infos mehr, da wir ja jetzt gerade gucken */
  chaninfo[AktChan]->flags &= ~CH_NEWDATA;

  /* Position des Cursors neu setzen */
  x_akt = chaninfo[AktChan]->x_akt;
  y_akt = chaninfo[AktChan]->y_akt;

  display_newscreen();
  
}



void send_all_awaytxt(char *text, char weg)
{
  char tmp[500],out[500];
  int chan;

  chan = 1;
  while (chan <= chancount)
  {
    if (chaninfo[chan]->driverfd != -1)
       if (chaninfo[chan]->type == TYPE_TERMINAL)
       {
         macro_thisline(tmp,text,chan);
         sprintf(out,"## %s ##\r",tmp);
         send_Textline(chan,out);
       }
       if (chaninfo[chan]->type == TYPE_CONVERS)
       {
         if (!weg)
            send_Textline(chan,"/a\r");
         else
         {
           macro_thisline(tmp,text,chan);
           sprintf(out,"/a %s\r",tmp);
           send_Textline(chan,out);
         }
       }
    chan++;
  }
}



/* Schreibt einen bestimmten Text als away-Text 'rein, oder meldet
 * den Operator wieder zurck
 */
void cmd_away(char *param)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  awaytime = time(NULL);
  if (param[0] == '\0')
  {
    if (awaytext == NULL)
    {
      send_all_awaytxt(config->awaytxt->weg,TRUE);
      awaytext = (char *) strdup(config->awaytxt->weg);
      ErrorText("You are marked as being away",TRUE);
    }
    else
    {
      send_all_awaytxt(config->awaytxt->da,FALSE);
      free(awaytext);
      awaytext = NULL;
      ErrorText("You are back again",TRUE);
    }
    return;
  }
  
  if (awaytext != NULL) free(awaytext);
  
  awaytext = (char *) strdup(param);
  send_all_awaytxt(awaytext,TRUE);
  ErrorText("You are marked as being away",TRUE);
}



void cmd_telefon()
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  awaytime = time(NULL);
  awaytext = (char *) strdup(config->awaytxt->telefon);
  send_all_awaytxt(awaytext,TRUE);
  ErrorText("You are marked as being away",TRUE);
}



void cmd_2m()
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);
  
  awaytime = time(NULL);
  awaytext = (char *) strdup(config->awaytxt->sleep);
  send_all_awaytxt(awaytext,TRUE);
  ErrorText("You are marked as being away",TRUE);
}



void cmd_monitor(char *params)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (params[0] == '\0')
  {
    ErrorText("Syntax: MONITOR ON|OFF",TRUE);
    return;
  }
  
  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    comm_monitor_activate();
    ErrorText("Ok",TRUE);
    return;
  }
  
  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    comm_monitor_deactivate();
    ErrorText("Ok",TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}



void cmd_clock(char *params)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (params[0] == '\0')
  {
    ErrorText("Syntax: CLOCK ON|OFF",TRUE);
    return;
  }
  
  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    config->clock = TRUE;
    ErrorText("Ok",TRUE);
    update_lower_statusline(TRUE);
    return;
  }

  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    config->clock = FALSE;
    ErrorText("Ok",TRUE);
    update_lower_statusline(TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}



/* Den angegebenen //-Befehl ausfhren */
void cmd_force(char *cmd)
{

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }

  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected.",TRUE);
    return;
  }

  remote_command(cmd,strlen(cmd),AktChan);
}



void cmd_klack(char *params)
{
  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (params[0] == '\0')
  {
    if (config->tones->klack)
       ErrorText("ON",TRUE);
    else
       ErrorText("OFF",TRUE);
    return;
  }
  
  Gross(params);
  
  if (!strcmp(params,"ON") || (params[0] == 'Y') || (params[0] == '1'))
  {
    config->tones->klack = TRUE;
    ErrorText("Ok",TRUE);
    return;
  }

  if (!strcmp(params,"OFF") || (params[0] == 'N') || (params[0] == '0'))
  {
    config->tones->klack = FALSE;
    ErrorText("Ok",TRUE);
    return;
  }
  
  ErrorText("Wrong parameter",TRUE);
}



/* Bricht eine AutoBIN- oder sonstige bertragung ab. */
void cmd_abort()
{
  s_sendqueue *tmp;
  char str[100];
  int chan;

  /* Nchste Zeile */
  ScrollDown_Vorschreib(FALSE);

  if (AktChan == 0)
  {
    ErrorText("Not in the monitor-screen.",TRUE);
    return;
  }
  
  if (chaninfo[AktChan]->driverfd == -1)
  {
    ErrorText("Only while connected!",TRUE);
    return;
  }

  chan = AktChan;

  if (chaninfo[chan]->sendqueue == NULL) return;

  switch (chaninfo[chan]->sendqueue->type)
  {
    case 3:
    case 4: /* Textfile-Send */
            tmp = chaninfo[chan]->sendqueue;
            chaninfo[chan]->sendqueue = chaninfo[chan]->sendqueue->next;
            free(tmp->plainfile->pfad);
            free(tmp->plainfile);
            free(tmp);
            strcpy(str,"\xD<GNOM>: Textfile-TX aborted by the SysOp.\xD");
            break;
    case 5: /* normales AutoBIN-File */
            close(chaninfo[chan]->sendqueue->autobin->fd);
            chaninfo[chan]->autobin = NULL;	/* Der Speicher wird in NextQueueEntry frei gemacht */
            chaninfo[chan]->flags &= ~CH_AUTOBINTX;
            
            tmp = chaninfo[chan]->sendqueue;
            chaninfo[chan]->sendqueue = chaninfo[chan]->sendqueue->next;
            free(tmp->autobin->name);
            free(tmp->autobin);
            free(tmp);

            /* #ABORT# schicken */
            send_Textline(chan,"#ABORT#\xD");
            
            strcpy(str,"\xD<GNOM>: AutoBIN-TX aborted by the SysOp.\xD");
            break;
  }
  
  if (chaninfo[chan]->type == TYPE_TERMINAL)
  {
    send_Textline(chan,str);
    send_a_frame(chan);
  }
  else display_addline(str,strlen(str),chan,ATTR_RX_SEND,TRUE);

}


