/*
 *  GNOM - packet radio-terminal for Linux
 *  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/****************************************************************************/
/**  Routinen fr die Sendqueue-Verwaltung                                 **/
/****************************************************************************/


#include "global.h"
#include "sndqueue.h"

#include "comp.h"
#include "comm.h"
#include "utils.h"
#include "display.h"
#include "convert.h"
#include "xabin.h"

#include <string.h>


/* Einen String endgltig aussenden */
static void sendit(int chan, char *str, int len)
{
  int i,complen;
  char compstr[500];
  
  /* Online-Komprimierung? Wenn ja -> komprimieren */
  if (((chaninfo[chan]->flags & CH_COMP) == CH_COMP) && ((chaninfo[chan]->sendqueue->type == 2) || 
                                                         (chaninfo[chan]->sendqueue->type == 4)))
  {
    i = encstathuf(str,len,compstr,&complen);
    if (i == 0)
       comm_send_frame(chan,compstr,complen);
    else
       comm_send_frame(chan,str,len);
  }
  else comm_send_frame(chan,str,len);
}



void send_switch_lfcr(char *tmp, int count)
{
  int i=0;
  
  while (i < count)
  {
    if (tmp[i] == '\n') tmp[i] = '\r';
    i++;
  }
}



/* Einen 0-terminierten String aussenden */
void send_Textline(int chan,char *str)
{
  s_sendqueue *tmp;

  /* Nur, wenn wir connected sind */
  if (chaninfo[chan]->driverfd == -1) return;
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Erster Eintrag */
    chaninfo[chan]->sendqueue = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = chaninfo[chan]->sendqueue;
  }
  else {
    tmp = chaninfo[chan]->sendqueue;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = tmp->next;
  }

  tmp->next = NULL;
  if ((chaninfo[chan]->flags & CH_COMP) == 0)
     tmp->type = 1;		/* Normaler Text */
  else
     tmp->type = 2;		/* Den Text komprimieren */
  tmp->plaintext = (s_plaintext *)malloc(sizeof(s_plaintext));
  tmp->plaintext->len = strlen(str);
  tmp->plaintext->text = (char *)strdup(str);

  if ((chaninfo[chan]->flags & CH_CAN_SEND) != 0)
     send_a_frame(chan);
}



void send_Bin(int chan, char *str, int len)
{
  s_sendqueue *tmp;
  
  /* Nur, wenn wir connected sind */
  if (chaninfo[chan]->driverfd == -1) return;
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Erster Eintrag */
    chaninfo[chan]->sendqueue = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = chaninfo[chan]->sendqueue;
  }
  else {
    tmp = chaninfo[chan]->sendqueue;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = tmp->next;
  }

  tmp->next = NULL;
  tmp->type = 7;
  tmp->plaintext = (s_plaintext *)malloc(sizeof(s_plaintext));
  tmp->plaintext->len = len;
  tmp->plaintext->text = (char *)malloc(len);
  memcpy(tmp->plaintext->text,str,len);

  if ((chaninfo[chan]->flags & CH_CAN_SEND) != 0)
     send_a_frame(chan);
}



/* Sendet ein File 1:1 ohne irgendwelche Wandlungen aus. */
void send_AutoBIN(int chan)
{
  s_sendqueue *tmp;
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Erster Eintrag */
    chaninfo[chan]->sendqueue = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = chaninfo[chan]->sendqueue;
  }
  else {
    tmp = chaninfo[chan]->sendqueue;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  tmp->type = 5;		/* AutoBIN-File abschicken */
  tmp->autobin = chaninfo[chan]->autobin;
  
  if ((chaninfo[chan]->flags & CH_CAN_SEND) != 0)
     send_a_frame(chan);
}



/* Sendet ein File mit dem neuen AutoBIN-Modus aus. */
void send_AutoEBIN(int chan)
{
  s_sendqueue *tmp;
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Erster Eintrag */
    chaninfo[chan]->sendqueue = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = chaninfo[chan]->sendqueue;
  }
  else {
    tmp = chaninfo[chan]->sendqueue;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  tmp->type = 6;		/* extendet AutoBIN-File abschicken */
  tmp->autobin = chaninfo[chan]->autobin;
  
  if ((chaninfo[chan]->flags & CH_CAN_SEND) != 0)
     send_a_frame(chan);
}



/* Ein File aussenden, \n -> \r - Umwandlung */
void send_Plainfile(int chan, char *pfad)
{
  s_sendqueue *tmp;
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Erster Eintrag */
    chaninfo[chan]->sendqueue = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = chaninfo[chan]->sendqueue;
  }
  else {
    tmp = chaninfo[chan]->sendqueue;
    while (tmp->next != NULL) tmp = tmp->next;
    
    tmp->next = (s_sendqueue *)malloc(sizeof(s_sendqueue));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  
  if ((chaninfo[chan]->flags & CH_COMP) == 0)
     tmp->type = 3;		/* Plain-File (1:1 aussenden) */
  else
     tmp->type = 4;		/* Plain-File (1:1 aussenden) - //COMP */
  tmp->plainfile = (s_plainfile *)malloc(sizeof(s_plainfile));
  tmp->plainfile->fd = -1;
  tmp->plainfile->pfad = (char *)strdup(pfad);

  if ((chaninfo[chan]->flags & CH_CAN_SEND) != 0)
     send_a_frame(chan);
}



/* Speicher fr den aktuellen Eintrag frei machen, und den nchsten vorholen */
void NextQueueEntry(int chan)
{
  s_sendqueue *tmp;
  
  tmp = chaninfo[chan]->sendqueue;
  chaninfo[chan]->sendqueue = chaninfo[chan]->sendqueue->next;
  
  switch (tmp->type)
  {
    case 1: 
    case 2: free(tmp->plaintext->text);
            free(tmp->plaintext);
            free(tmp);
            break;
    case 3: 
    case 4: free(tmp->plainfile->pfad);
            free(tmp->plainfile);
            free(tmp);
            break;
    case 5: free(tmp->autobin->name);
            free(tmp->autobin);
            free(tmp);
            break;
    case 6: free(tmp);
            break;
    case 7: free(tmp->plaintext->text);
            free(tmp->plaintext);
            free(tmp);
            break;
  }
}



/* Verschickt das nchste Frame auf diesem Kanal.
 */
void send_a_frame(int chan)
{
  int len,count,ready,delete_it=FALSE;
  char *tmp,tmp2[500],*timeptr;
  time_t timediff;
  
  tmp = (char *)malloc((framesize*2)+1);
  
  if (chaninfo[chan]->sendqueue == NULL)
  {
    /* Wurde alles gesendet? Ja -> gucken, ob disconnected werden soll */
    if ((chaninfo[chan]->flags & CH_DISC) == CH_DISC) disc_channel(chan);
    free(tmp);
    return;
  }
  
  /* Nach diesem Frame automatisch noch nix senden */
  chaninfo[chan]->flags &= ~CH_CAN_SEND;
  
  
  switch (chaninfo[chan]->sendqueue->type)
  {
    case 1: /* Normalen Text oder... */
    case 2: /* ...komprimierten Text aussenden */
            len = framesize;
            if (chaninfo[chan]->sendqueue->plaintext->len < len) len = chaninfo[chan]->sendqueue->plaintext->len;
            
            /* Zu sendende Daten nach tmp kopieren */
            memcpy(tmp,chaninfo[chan]->sendqueue->plaintext->text,len);
            tmp[len] = '\0';
            
            /* Wenn noch was von dem Text brig ist, wird er wieder in die Queue zurck
             * kopiert, ansonsten das nchste Queue-Element holen
             */
            if (chaninfo[chan]->sendqueue->plaintext->len > len)
            {
              memmove(chaninfo[chan]->sendqueue->plaintext->text,
                      chaninfo[chan]->sendqueue->plaintext->text + framesize,
                      chaninfo[chan]->sendqueue->plaintext->len - len);
              chaninfo[chan]->sendqueue->plaintext->len = chaninfo[chan]->sendqueue->plaintext->len - len;
            }
            else delete_it = TRUE;
            
            /* Frame evtl.  convertieren */
            conv_it_tx(tmp,len,chaninfo[chan]->convert_tx);
            
            /* Frame Aussenden */
            sendit(chan,tmp,len);
            
            /* und Anzeigen */
            checkshowchars(tmp,len);
            display_addline(tmp,len,chan,ATTR_RX_SEND,TRUE);
            
            /* Wird gerade etwas gespeichert? */
            if ((chaninfo[chan]->flags & CH_SAVE) == CH_SAVE)
            {
              if (chaninfo[chan]->saveplain->local == 1)
              {
                write(chaninfo[chan]->saveplain->fd,tmp,len);
                write(chaninfo[chan]->saveplain->fd,"\xA",1);
              }
            }
            free(tmp);
            
            if (delete_it) NextQueueEntry(chan);
            return;
    
    
    case 3: /* Ein File aussenden, \n -> \r - Umwandlung */
    case 4: /* komprimiert */
            if (chaninfo[chan]->sendqueue->plainfile->fd == -1)
            {
              /* das File ist noch zu - erstmal ffnen */
              chaninfo[chan]->sendqueue->plainfile->fd =
                 open(chaninfo[chan]->sendqueue->plainfile->pfad,O_RDONLY);
            }
            if (chaninfo[chan]->sendqueue->plainfile->fd != -1)
            {
              /* OK, ffnen hat jetzt einwandfrei geklappt */
              ready = FALSE;
              
              count = read(chaninfo[chan]->sendqueue->plainfile->fd,tmp,framesize);
              tmp[count] = '\0';
              if (count != framesize) ready = TRUE;
              
              /* Umwandlung \n -> \r */
              send_switch_lfcr(tmp,count);
              
              /* Frame senden */
              sendit(chan,tmp,count);
/*              comm_send_frame(chan,tmp,count);*/
              
              /* Und Anzeigen */
              checkshowchars(tmp,count);
              display_addline(tmp,count,chan,ATTR_RX_SEND,TRUE);
              
              /* Fertig mit der Ausgabe? */
              if (ready)
              {
                close(chaninfo[chan]->sendqueue->plainfile->fd);
                NextQueueEntry(chan);
              }
            }
            else NextQueueEntry(chan);
            free(tmp);
            return;
            
    case 5: /* Ein AutoBIN-File aussenden */
            if (chaninfo[chan]->sendqueue->autobin->fd == -1)
            {
              /* das File ist noch zu - erstmal ffnen */
              chaninfo[chan]->sendqueue->autobin->fd =
                 open(chaninfo[chan]->sendqueue->autobin->name,O_RDONLY);
                      
              chaninfo[chan]->sendqueue->autobin->starttime = time(NULL);
              chaninfo[chan]->flags |= CH_AUTOBINTX;
            }
            if (chaninfo[chan]->sendqueue->autobin->fd != -1)
            {
              /* OK, ffnen hat jetzt einwandfrei geklappt */
              ready = FALSE;
              
              count = read(chaninfo[chan]->sendqueue->autobin->fd,tmp2,framesize);
              chaninfo[chan]->sendqueue->autobin->bytes_got += framesize;
              if (count != framesize) ready = TRUE;
              update_lower_statusline(TRUE);
              
              /* Frame senden */
              comm_send_frame(chan,tmp2,count);
              
              /* Fertig mit der Ausgabe? */
              if (ready)
              {
                timediff = time(NULL) - chaninfo[chan]->sendqueue->autobin->starttime;
                if (timediff == 0) timediff = 1;
                timeptr = (char *)spec_time(timediff);
                
                sprintf(tmp2,"<GNOM>: AutoBIN-TX OK. CRC: %i (time: %s, %li baud)\xD",
                             chaninfo[chan]->sendqueue->autobin->crc_tx,timeptr,
                             (chaninfo[chan]->sendqueue->autobin->len*8)/timediff);
                free(timeptr);
                
                close(chaninfo[chan]->sendqueue->autobin->fd);
                chaninfo[chan]->autobin = NULL;	/* Der Speicher wird in NextQueueEntry frei gemacht */
                chaninfo[chan]->flags &= ~CH_AUTOBINTX;
                NextQueueEntry(chan);
                
                if (chaninfo[chan]->type == TYPE_TERMINAL)
                {
                  send_Textline(chan,tmp2);
                  send_a_frame(chan);
                }
                else
                   display_addline(tmp2,strlen(tmp2),chan,ATTR_RX_SEND,TRUE);
                free(tmp);
                return;
              }
            }
            else NextQueueEntry(chan);
            free(tmp);
            return;
    case 6: /* extendet AutoBIN-File aussenden */
            if (chaninfo[chan]->sendqueue->autobin->fd == -1)
            {
              /* das File ist noch zu - erstmal ffnen */
              chaninfo[chan]->sendqueue->autobin->fd =
                 open(chaninfo[chan]->sendqueue->autobin->name,O_RDONLY);
              
              lseek(chaninfo[chan]->sendqueue->autobin->fd,chaninfo[chan]->AutoEBINoffset,SEEK_SET);
              
              chaninfo[chan]->sendqueue->autobin->starttime = time(NULL);
              chaninfo[chan]->flags |= CH_AUTOBINTX;
              chaninfo[chan]->flags &= ~CH_LINEMODE;
              chaninfo[chan]->AutoEBIN = NULL;
              chaninfo[chan]->AutoEBINblocksize = 8;
              chaninfo[chan]->AutoEBINoffset = 0;
              chaninfo[chan]->autobin = chaninfo[chan]->sendqueue->autobin;
              chaninfo[chan]->AutoEBINready = FALSE;
              chaninfo[chan]->ebin = NULL;
            }
            
            if (chaninfo[chan]->sendqueue->autobin->fd != -1)
            {
              if (chaninfo[chan]->AutoEBIN == NULL)
                 ebin_get_block(chan,&chaninfo[chan]->sendqueue->autobin->fd);
              
              len = framesize;
              if (chaninfo[chan]->AutoEBINlen < len) len = chaninfo[chan]->AutoEBINlen;
              chaninfo[chan]->sendqueue->autobin->bytes_got += len;
              
              /* Zu sendende Daten nach tmp kopieren */
              memcpy(tmp,chaninfo[chan]->AutoEBIN,len);
              tmp[len] = '\0';
              
              
              /* Frame senden */
              comm_send_frame(chan,tmp,len);
              
              /* Wenn noch was von dem Text brig ist, wird er wieder in die Queue zurck
               * kopiert, ansonsten das nchste Queue-Element holen
               */
              if (chaninfo[chan]->AutoEBINlen > len)
              {
                memmove(chaninfo[chan]->AutoEBIN,
                        chaninfo[chan]->AutoEBIN + framesize,
                        chaninfo[chan]->AutoEBINlen - len);
                chaninfo[chan]->AutoEBINlen = chaninfo[chan]->AutoEBINlen - len;
              }
              else
              {
                free(chaninfo[chan]->AutoEBIN);
                chaninfo[chan]->AutoEBIN = NULL;
              }
            }
            
            if (chaninfo[chan]->AutoEBINready)
            {
              /* File ist zuende */
              timediff = time(NULL) - chaninfo[chan]->sendqueue->autobin->starttime;
              if (timediff == 0) timediff = 1;
              timeptr = (char *)spec_time(timediff);
              
              sprintf(tmp2,"\xD<GNOM>: extended AutoBIN-TX OK. CRC=%i (time: %s, %li baud)\xD",
                           chaninfo[chan]->sendqueue->autobin->crc_tx,timeptr,
                           (chaninfo[chan]->sendqueue->autobin->len*8)/timediff);
              free(timeptr);
              
/*              chaninfo[chan]->flags &= ~CH_AUTOBINTX;*/
              
              NextQueueEntry(chan);
              
/*              if (chaninfo[chan]->type == TYPE_TERMINAL)
              {
                send_Textline(chan,tmp2);
                send_a_frame(chan);
              }
              else display_addline(tmp2,strlen(tmp2),chan,ATTR_RX_SEND,TRUE);*/
            }
            free(tmp);
            return;
    case 7: /* Einen Text 1:1 (z.B. Binr) aussenden */
            len = framesize;
            if (chaninfo[chan]->sendqueue->plaintext->len < len) len = chaninfo[chan]->sendqueue->plaintext->len;
            
            /* Zu sendende Daten nach tmp kopieren */
            memcpy(tmp,chaninfo[chan]->sendqueue->plaintext->text,len);
            tmp[len] = '\0';
            
            /* Wenn noch was von dem Text brig ist, wird er wieder in die Queue zurck
             * kopiert, ansonsten das nchste Queue-Element holen
             */
            if (chaninfo[chan]->sendqueue->plaintext->len > len)
            {
              memmove(chaninfo[chan]->sendqueue->plaintext->text,
                      chaninfo[chan]->sendqueue->plaintext->text + framesize,
                      chaninfo[chan]->sendqueue->plaintext->len - len);
              chaninfo[chan]->sendqueue->plaintext->len = chaninfo[chan]->sendqueue->plaintext->len - len;
            }
            else delete_it = TRUE;
            
            /* Frame Aussenden */
            sendit(chan,tmp,len);
            
            /* und Anzeigen */
            if ((chaninfo[chan]->flags & CH_AUTOBINTX) == 0)
            {
              checkshowchars(tmp,len);
              display_addline(tmp,len,chan,ATTR_RX_SEND,TRUE);
            }
            
            /* Wird gerade etwas gespeichert? */
            if ((chaninfo[chan]->flags & CH_SAVE) == CH_SAVE)
            {
              if (chaninfo[chan]->saveplain->local == 1)
              {
                write(chaninfo[chan]->saveplain->fd,tmp,len);
                write(chaninfo[chan]->saveplain->fd,"\xA",1);
              }
            }
            free(tmp);
            
            if (delete_it) NextQueueEntry(chan);
            return;
  }
}


